/*  Last edited: Jan  4 20:48 1997 (badger) */
/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */
/*  Last edited: Oct  4 15:00 1996 (rmott) */
/* $Id: cafcat.c,v 1.1 1997/04/03 13:44:09 badger Exp $ */
/* $Log: cafcat.c,v $
 * Revision 1.1  1997/04/03  13:44:09  badger
 * Initial revision
 *
 * Revision 1.3  1996/10/04  13:22:02  rmott
 * nothing
 *
 * Revision 1.2  1996/10/03  17:22:29  rmott
 * *** empty log message ***
 *
 * Revision 1.1  1996/03/08  10:06:56  rmott
 * Initial revision
 * */

#define MAIN 1

#include<stdio.h>
#include"cl.h"
#include"caf.h"
#include"scf.h"

int
main(int argc, char **argv)
{
  DICT *seqDict;
  Array seqs;
  CAFSEQ *seq, *s;
  FILE *fp, *out_fp, *fofn;
  int i, j, k, len;
  ASSINF *a;
  Scf *scf;
  cafAssembly *CAF;
  char CAFfile[256];
  char outfile[256];
  char fofnfile[256];
  char *dna;
  int reformat=0;
  int verbose=0;
  int check=0;
  int summary=0;
  int sanitise=1;

  strcpy(CAFfile,"stdin");
  strcpy(outfile,"stdout");
  
  getboolean("-reformat",&reformat,argc,argv);
  getboolean("-verbose",&verbose,argc,argv);
  getboolean("-check",&check,argc,argv);
  getboolean("-summary",&summary,argc,argv);
  getboolean("-sanitise",&sanitise,argc,argv);

  fofn = argfile("-fofn","r",argc,argv,fofnfile); /* use just the sequences listed in fofn - otherwise use all */
  CAF = readCAF( argc, argv, CAFfile );
  getarg("-out",outfile,argc,argv);
  
  gethelp(argc,argv);
  
  if ( CAF )
    {
      if ( reformat )
	{
	  seqs = CAF->seqs;
	  seqDict = CAF->seqDict;
	  for(i=0;i<arrayMax(seqs);i++)
	    {
	      seq = arrp(seqs,i,CAFSEQ);
	      if ( seq->SCF_File )
		seq->type = is_read;
	      if ( seq->dna )
		{
		  dna = stackText(seq->dna,0);
		  len = strlen(dna);
		  for(j=0;j<len;j++)
		    {
		      if ( dna[j] == '-' )
			dna[j] = 'n';
		      else if ( dna[j] == '*' )
			dna[j] = '-';
		    }
		}
	    }
	}


/*      
      {
	CAFSEQ *assembly;
	CAFSEQ *group;
	CAFSEQ *contig;
	GROUP_ORDER *go, *co;

	seqs = CAF->seqs;
	seqDict = CAF->seqDict;

	for(i=0;i<arrayMax(seqs);i++)
	  {
	    assembly = arrp(seqs,i,CAFSEQ);
	    if ( is_assembly == assembly->type )
	      {
		fprintf( stderr, "Assembly %s\n", dictName(seqDict, assembly->id ) );
		if ( assembly->order ) 
		  {
		    for(j=0;j<arrayMax(assembly->order);j++) 
		      {
			go = arrp(assembly->order,j,GROUP_ORDER);
			fprintf(stderr, "   group %s %g\n", dictName(seqDict,go->seq), go->position);
			group = arrp(seqs,go->seq,CAFSEQ);
			if ( group->order ) 
			  {
			    for(k=0;k<arrayMax(group->order);k++)
			      {
				co = arrp(group->order,k,GROUP_ORDER);
				contig = arrp(seqs,co->seq,CAFSEQ);
				fprintf(stderr, "           contig %s %g %s %d\n", dictName(seqDict, co->seq), co->position, sequence_text[contig->type], contig->id );
			      }
			  }
		      }
		  }
	      }
	  }
	
      }

*/
      if ( summary )
	{
	  summariseCAF( CAF );
	  exit(0);
	}
      
      
      if ( check )
	exit( check_holes( CAF, verbose ) + checkCAF( CAF, verbose ) );
      
      
      if ( fofn )
	{
	  char read[256];
	  char line[256];
	  Array write_seq;
	  CAFSEQ *contig;
	  int j, nseq=0, gseq=0;

	  seqs = CAF->seqs;
	  seqDict = CAF->seqDict;
	  write_seq = arrayCreate(arrayMax(seqs),int);
	  for(i=0;i<arrayMax(seqs);i++)
	    *arrayp(write_seq,i,int) = 0;

	  fprintf(stderr, "Only writing sequences in %s...\n", fofnfile );
	  while(skip_comments(fofn,line)!= EOF )
	    {
	      sscanf(line,"%s",read );
	      if ( dictFind(seqDict,read,&i) )
		{
		  (*arrayp(write_seq,i,int))++;
		  contig = arrp(seqs,i,CAFSEQ);
		  if ( is_contig == contig->type ) /* select all reads in the contig too */
		    {
		      for(j=0;j<arrayMax(contig->assinf);j++)
			{
			  a = arrp(contig->assinf,j,ASSINF);
			  (*arrayp(write_seq,a->seq,int))++;
			}
		    }
		}  
	      else
		fprintf(stderr,"Unknown Sequence: %s\n", read );
	    }

	  for(i=0;i<arrayMax(seqs);i++)
	    {
	      seq = arrp( seqs, i, CAFSEQ );
	      nseq++;
	      if ( array(write_seq,i,int) == 0 )
		seq->bad = TRUE;
	      else
		{
		  seq->bad = FALSE;
		  gseq++;
		}
	    }
	  fprintf(stderr, "%d/%d Sequences selected\n", gseq, nseq );
	}
      
      if ( sanitise )
	sanitiseCAF(CAF);

      writeCAF( CAF, CAFfile, argc, argv );
      
      if ( verbose )
	{
	  seqs = CAF->seqs;
	  seqDict = CAF->seqDict;
	  for(i=0;i<arrayMax(seqs);i++)
	    {
	      seq = arrp( seqs, i, CAFSEQ );
	      if ( seq->type == is_contig && seq->assinf )
		{
		  printf("\n\n ***** %s ***** \n\n", dictName(seqDict, i ) );
		  for(j=0;j < arrayMax(seq->assinf); j++)
		    {
		      a = arrayp(seq->assinf, j, ASSINF) ;
		      fprintf (stdout, "Assembled_from %s %d %d %d %d\n", 
			       dictName (seqDict, a->seq), a->s1, a->s2, a->r1, a->r2) ;
		      a->s = ass_subseq( seqs, a );
		      fprintf(stdout,"%s\n", a->s ); 
		    }
		}
	      if ( seq->SCF_File && seq->assinf )
		{
		  printf("\n\n ***** %s ***** SCF \n\n", dictName(seqDict, i ) );
		  if ( scf = get_scf(seq->SCF_File) )
		    {
		      for(j=0;j < arrayMax(seq->assinf); j++)
			{
			  a = arrayp(seq->assinf, j, ASSINF) ;
			  fprintf (stdout, "Assembled_from %s %d %d %d %d\n", 
				   dictName (seqDict, a->seq), a->s1, a->s2, a->r1, a->r2) ;
			}
		    }
		}
	    }
	}
      return 0;
    }
  return 1;
}
