/*  Last edited: Jan  4 20:48 1997 (badger) */
/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */
#include<stdio.h>
#include<string.h>
#include"io_buffer.h"

/* Functions for buffered unformatted IO.

 Idea is to be able to read and write single arbitrary objects, whilst
 avoiding the overhead of many reads and writes
 */


IO_BUFFER *
create_buffer( size_t max, size_t size_of )
{
  IO_BUFFER *B = (IO_BUFFER*)calloc(1,sizeof(IO_BUFFER));
  B->buffer = (void*)calloc(max, size_of);
  B->size = 0;
  B->max = max;
  B->size_of = size_of;

 return B;
}

void
destroy_buffer ( IO_BUFFER *B )
{
  if ( B )
    {
      if ( B->buffer ) 
	free(B->buffer);
      free(B);
    }
}
 
int 
buffered_fread( void *address, IO_BUFFER *B, FILE *fp )
{
/* Returns one item */

  if ( B->size > 0 ) /* return the right-most item */
    {
      char *b = (char*)B->buffer;
      B->size--;
      memcpy(address,b+B->size*B->size_of,B->size_of);
    }
  else
    {
/* read in the next block */

      if ( (B->size = fread( B->buffer, B->size_of, B->max, fp)) <= 0 )
	return 0;
      else
	{
	  char *b = (char*)B->buffer;
	  /* reverse the order of items */
	  {
	    char *tmp = (char*)calloc(1,B->size_of);
	    int n, m;
	    char *bn, *bm;
	    for(n=0,m=B->size-1;n<B->size/2;n++,m--)
	      {
		bn = b+n*B->size_of;
		bm = b+m*B->size_of;
		memcpy(tmp,bn,B->size_of);
		memcpy(bn,bm,B->size_of);
		memcpy(bm,tmp,B->size_of);
	      }
	    free(tmp);
	  }
	  B->size--;
	  memcpy(address,b+B->size*B->size_of,B->size_of);
	}
    }
  return 1;
}

int 
buffered_fwrite( void *address, IO_BUFFER *B, FILE *fp )
{
  size_t s;

  if ( address == NULL || B->size == B->max ) /* flush */
    {
      s = fwrite( B->buffer, B->size_of, B->size, fp );
      B->size = 0;
    }
  if ( address )
    {
      char *b = (char*)B->buffer;
      memcpy( b+B->size*B->size_of, address, B->size_of );
      B->size++;
      s = 1;
    }

  return s;
}

  



