/*  Last edited: Jan  4 20:48 1997 (badger) */
/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */
#include<stdio.h>
#include<ctype.h>
#include"iub.h"

/* support for sequence matching using IUB ambiguity codes */

int IUBunequal(x,y)

char x, y;
{
  static char iub[256][256];
  static int initialised=0;

  if ( ! initialised )
    {
      char A = 'A';
      char C = 'C';
      char G = 'G';
      char T = 'T';
      char U = 'U';
      char Y = 'Y';
      char M = 'M';
      char K = 'K';
      char S = 'S';
      char W = 'W';
      char H = 'H';
      char B = 'B';
      char V = 'V';
      char D = 'D';
      char N = 'N';
      char R = 'R';
      int i, j;

      iub[A][A] = A;
      iub[C][C] = C;
      iub[G][G] = G;
      iub[T][T] = T;

      iub[R][R] = R;
      iub[R][A] = A;
      iub[R][G] = G;

      iub[R][Y] = 0;
      iub[R][M] = A;
      iub[R][K] = G;
      iub[R][S] = G;
      iub[R][W] = A;
      iub[R][H] = A;
      iub[R][B] = G;
      iub[R][V] = R;
      iub[R][D] = R;
      iub[R][N] = R;

      iub[Y][C] = C;
      iub[Y][T] = T;
      iub[Y][Y] = Y;

      iub[Y][U] = U;
      iub[Y][M] = C;
      iub[Y][K] = T;
      iub[Y][S] = C;
      iub[Y][W] = T;
      iub[Y][H] = C;
      iub[Y][B] = Y;
      iub[Y][V] = C;
      iub[Y][D] = T;
      iub[Y][N] = Y;

      iub[M][A] = A;
      iub[M][C] = C;
      iub[M][M] = M;

      iub[M][S] = C;
      iub[M][W] = A;
      iub[M][H] = M;
      iub[M][B] = C;
      iub[M][V] = M;
      iub[M][D] = A;
      iub[M][N] = M;
      
      iub[K][G] = G;
      iub[K][T] = T;
      iub[K][U] = U;

      iub[S][C] = C;
      iub[S][G] = G;
      iub[S][S] = S;

      iub[S][H] = C;
      iub[S][B] = S;
      iub[S][V] = S;
      iub[S][D] = G;
      iub[S][N] = S;
      
      iub[W][A] = A;
      iub[W][T] = T;
      iub[W][U] = U;

      iub[W][W] = W;
      iub[W][H] = W;
      iub[W][B] = T;
      iub[W][V] = A;
      iub[W][D] = W;
      iub[W][N] = W;

      iub[B][C] = C;
      iub[B][G] = G;
      iub[B][T] = T;
      iub[B][U] = U;

      iub[H][A] = A;
      iub[H][C] = C;
      iub[H][T] = T;
      iub[H][U] = U;

      iub[H][H] = H;
      iub[H][B] = Y;
      iub[H][V] = M;
      iub[H][D] = W;
      iub[H][N] = N;

      iub[V][A] = A;
      iub[V][C] = C;
      iub[V][G] = G;

      iub[V][V] = V;
      iub[V][D] = R;
      iub[V][N] = V;
      
      iub[D][A] = A;
      iub[D][G] = G;
      iub[D][T] = T;
      iub[D][U] = U;

      iub[D][D] = D;
      iub[D][N] = D;

      iub[N][A] = A;
      iub[N][C] = C;
      iub[N][G] = G;
      iub[N][T] = T;

      iub[N][N] = N;
      
      for(i=0;i<256;i++)
	for(j=0;j<i;j++)
	  {
	    if ( ! iub[i][j] )
	      iub[i][j] = iub[j][i];
	    else
	      iub[j][i] = iub[i][j];
	    if ( iub[i][j] )
	      printf("%c %c = %c\n", i, j, iub[i][j] );
	  }

      initialised = 1;
    }
  
  return iub[toupper(x)][toupper(y)];
}

int IUBcmp(a,b)

char *a, *b;
{
  int n;

  while( *a && *b )
    {
      n=IUBunequal(*a++,*b++);
      if ( n )
	return n;
    }
  if ( *a ) 
    return 1;
  else if ( *b )
    return -1;
  else
    return 0;
}
      
	
