/*  Last edited: Jan  4 20:48 1997 (badger) */
/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */
/*  Last edited: Aug 15 17:39 1995 (rmott) */
#define PRIME_NUM 10007
#define MAX_BUF 10000
#define SEQPATH "SEQ_DATA_HOME"

#ifndef _SEQ_UTIL_H_
#define _SEQ_UTIL_H_

typedef enum 
{ EMBL, NBRF, FASTA, SINGLE  } 
WHAT_KIND_OF_DATABASE;

typedef struct hash_list
{
  char *name;
  unsigned long offset, text_offset;
  struct hash_list *next;
}
HASH_LIST;

typedef struct
{
  char *database;
  WHAT_KIND_OF_DATABASE type;
  int sequences;
  unsigned long length;
  FILE *datafile;
  FILE *textfile;
  FILE *indexfile;
  HASH_LIST **index;
}
DATABASE;


typedef struct
{
  char *name;
  DATABASE *database;
  char *desc;
  int len;
  char *s;
}
SEQUENCE;


SEQUENCE *read_sequence(), *read_sq(), *read_seq();
SEQUENCE *read_embl_sq(), *read_nbrf_sq(), *read_fasta_sq();
SEQUENCE *next_seq_from_list_file(), *next_seq_from_database_spec();
SEQUENCE *next_seq();
SEQUENCE *get_seq_from_file();
SEQUENCE *seqdup();
SEQUENCE *get_fasta_sq( FILE *fp );
SEQUENCE *shuffle_seq( SEQUENCE *seq, int in_place, int *seed );

char *shuffle_s( char *s, int *seed );

DATABASE *open_database(), *which_database(), *is_sequence_spec();
DATABASE *open_embl_database(), *open_nbrf_database(), *open_fasta_database();

FILE *which_file_of_sequences();

unsigned long get_offset(), seekto(), find_next();
char *downcase(), *upcase(), *complement_seq(), *clean_line();
char *seq_comment(), *embl_seq_comment(), *nbrf_seq_comment(), *fasta_seq_comment();

static DATABASE *databases[100];
static int database_count;

char *iubtoregexp();
char *iub2regexp();
char *iub_regexp( char c);




FILE *openfile_in_seqpath();
FILE *openfile_in_seqpath_arg();

SEQUENCE *into_sequence( char *name, char *desc, char *s );
SEQUENCE *subseq( SEQUENCE *seq, int start, int stop );
void free_seq( SEQUENCE *seq );

void make_embl_index( DATABASE *db );
void make_fasta_index( DATABASE *db );
void make_nbrf_index( DATABASE *db );


#endif
