/*  Last edited: Jan  4 20:48 1997 (badger) */
/*
 * This software has been created by Genome Research Limited (GRL). 
 * GRL hereby grants permission to use, copy, modify and distribute 
 * this software and its documentation for non-commercial purposes 
 * without fee at the user's own risk on the basis set out below.
 * GRL neither undertakes nor accepts any duty whether contractual or 
 * otherwise in connection with the software, its use or the use of 
 * any derivative, and makes no representations or warranties, express
 * or implied, concerning the software, its suitability, fitness for
 * a particular purpose or non-infringement.
 * In no event shall the authors of the software or GRL be 
 * responsible or liable for any loss or damage whatsoever arising in
 * any way directly or indirectly out of the use of this software or 
 *  its derivatives, even if advised of the possibility of such damage.
 * Our software can be freely distributed under the conditions set 
 * out above, and must contain this copyright notice.
 */
/*  Last edited: Jul 18 14:57 1996 (rmott) */
/* $Id: svec_clip.c,v 1.1 1997/04/03 13:44:09 badger Exp $ */
/* $Log: svec_clip.c,v $
 * Revision 1.1  1997/04/03  13:44:09  badger
 * Initial revision
 *
 * Revision 1.1  1996/10/03  17:23:55  rmott
 * Initial revision
 * */

#define MAIN 1

#include<stdio.h>
#include<string.h>
#include<math.h>
#include"caf.h"
#include"trace_align.h"
#include"seq_util.h"
#include"scf.h"
#include"cl.h"

void 
end_align( SEQUENCE *query, SEQUENCE *seq, 
	  int match, int mismatch, int gap, int max, int smith_waterman, 
	  int *q, int *s, int *score,  int *matches );

extern int sim_mat[256][256];


SEQUENCE *get_3_vec( SEQUENCE *vec, int sc, int sp );
SEQUENCE *get_5_vec( SEQUENCE *vec, int sc, int sp );
SEQUENCE *scf2seq( Scf *scf, char *name ); /* extract the sequence from an scf */

void 
try_trace_alignment( Scf *scf, SEQUENCE *vec, int argc, char **argv, 
			 int *clip_point, int *score, float *mean, float *var, int *ok, int *shuffle );
void 
try_seq_alignment( SEQUENCE *seq, SEQUENCE *vec, int max_clip, int argc, char **argv, 
		       int *clip_pos, int *clip_pos_vec, int *score, int *matches, 
		       float *mean, float *var, int *ok, int *shuffle );

int 
main(int argc,char **argv)
{
  SEQUENCE *query, *seq, *vec_3, *vec_5;
  FILE *query_fp, *seq_fp, *ps;
  char queryfile[256], seqfile[256], scf_file[256], ps_file[256];
  Scf *scf=NULL;
  int q, s, score, matches, i;
  int sp=0, sc=0;
  int both=1; /* do both types of screening */
  float mean, var;
  int ok, shuffle=20;
  int clip_vector;
  int clip_seq;
  char screenfile[256];
  FILE *sfp;
  char *t;
  t_params params;
  int max_clip=70;
  int found_3=FALSE, found_5=FALSE;

  getint("-sc",&sc,argc,argv); /* cloning site in vector */
  getint("-sp",&sp,argc,argv); /* primer offset in vector */
  getint("-shuffle", &shuffle,argc,argv); /* no of shuffles */

  if ( ! sp )
    {
      fprintf(stderr,"cloning site sc not defined\n");
      exit(1);
    }

  if ( ! sc )
    {
      fprintf(stderr,"primer site sp not defined\n");
      exit(1);
    }

  getint("-max",&max_clip,argc,argv);

  if ( sp && sc &&
      ( query_fp = argfile("-query","r",argc,argv,queryfile) ) && 
      ( query = get_fasta_sq( query_fp) ) &&
      ( (seq_fp = argfile("-seq","r",argc,argv,seqfile)) || 
	 ( getarg("-scf",scf_file,argc,argv) && ( scf = get_scf( scf_file ))))
      )
    {
      if ( scf ) /* trace alignment */
	{
	  seq = scf2seq( scf, scf_file); /* extract the sequence from the scf */

	  /* (a) 5' end */

	  vec_5 = get_5_vec( query, sc, sp );

	  /* first try the sequence alignment */
	  /* we allow shuffling */

	  try_seq_alignment( seq, vec_5, max_clip, argc, argv, &clip_seq, &clip_vector, &score, &matches, &mean, &var, &ok, &shuffle );

	  if ( ok  && ok >= shuffle ) /* found the vector */
	    {
	      printf("SL %-15s %5d %-15s %5d %5d %5d   ", vec_5->name, clip_vector, seq->name, clip_seq, score, matches );
	      printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, var, ((float)score-mean)/(var+1.0e-10), ok, shuffle );
	      t=seq->s;
	      while( t < &seq->s[clip_seq] )
		putc(*(t++),stdout);
	      putc(' ',stdout);
	      printf("\n");
	      found_5 = TRUE;
	    }
	  else /* try the trace alignment */
	    {
	      printf("! SL %-15s %5d %-15s %5d %5d %5d   ", vec_5->name, clip_vector, seq->name, clip_seq, score, matches );
	      printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, var, (score-mean)/(var+1.0e-10), ok, shuffle );
	      try_trace_alignment( scf, vec_5, argc, argv, &clip_seq, &score, &mean, &var, &ok, &shuffle );
	      printf("SL %-15s %5d %-15s %5d %5d %5d ", vec_5->name, vec_5->len, scf_file, clip_seq, score, 0 );
	      printf("    %5.2f %5.2f %6.3f   %2d / %2d trace aligned \n", mean, var, (score-mean)/(var+1.0e-10), ok, shuffle );
	    }

	  /* (b) 3' end */
	  /* THe 3' end is treated differently, mostly for reasons of
             efficiency. We just make one sequence alignment (with NO
             shuffles). THis should be ok because if sequencing vector
             is present at the 3'end it will be obvious */

	  vec_3 = get_3_vec( query, sc, seq->len );

	  complement_seq(seq->s); 
	  shuffle = 0;
	  try_seq_alignment( seq, vec_3, seq->len, argc, argv, &clip_seq, &clip_vector, &score, &matches, &mean, &var, &ok, &shuffle );
	  if ( ok   && ok >= shuffle )   
	    {/* found the vector */
	      printf("SR %-15s %5d %-15s %5d %5d %5d   ", vec_3->name, vec_3->len-clip_vector, seq->name, seq->len-clip_seq+1, score, matches );
	      printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, var, (score-mean)/(var+1.0e-10), ok, shuffle );
	      t=seq->s;
	      while( t < &seq->s[clip_seq] )
		putc(*(t++),stdout);
	      putc(' ',stdout);
	      printf("\n");
	      found_3 = TRUE;
	      
	    }
	  if ( ! found_3 && ! found_5 ) /* neither end found - test for completely vector */
	    {
	      int match=3, mismatch=-5, gap=-5, neutral=0;

	      getint("-match=%d",&match,argc,argv);
	      getint("-mismatch=%d",&mismatch,argc,argv);
	      getint("-gap=%d",&gap,argc,argv);
  
	      if ( mismatch  > 0 )
		mismatch = -mismatch;
	      if ( gap > 0 )
		gap = -gap;

	      score = smith_waterman_score2( query->s, seq->s, query->len, seq->len, match, mismatch, gap, neutral, 0, -1, &clip_vector, &clip_seq );
	      if ( score > 0.3*match*seq->len ) /* completely vector (match extands > 30% of the total read length) */
		{
		  printf("CC Completely Sequencing vector (svec_clip)\n");
		  printf("PS Completely Sequencing vector\n");
		}
	    }
	  free_seq(seq);
	}
      else /* sequence alignment */
	{
	  vec_5 = get_5_vec( query, sc, sp );
	  vec_3 = get_3_vec( query, sc, query->len );
	  while ( seq = get_fasta_sq( seq_fp ) )
	    {
	      try_seq_alignment( seq, vec_5, max_clip, argc, argv, &s, &clip_vector, &score, &matches, &mean, &var, &ok, &shuffle );
	      printf("SL %-15s %5d %-15s %5d %5d %5d   ", vec_5->name, clip_vector, seq->name, s, score, matches );
	      printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, var, (score-mean)/(var+1.0e-10), ok, shuffle );
	      t=seq->s;
	      while( t < &seq->s[s] )
		putc(*(t++),stdout);
	      putc(' ',stdout);
	      printf("\n");


	      complement_seq(seq->s);
	      
	      try_seq_alignment( seq, vec_3, max_clip, argc, argv, &clip_seq, &clip_vector, &score, &matches, &mean, &var, &ok, &shuffle );
	      printf("SR %-15s %5d %-15s %5d %5d %5d   ", vec_3->name, vec_3->len-clip_vector, seq->name, seq->len-clip_seq, score, matches );
	      printf("    %5.2f %5.2f %6.3f   %2d / %2d  ", mean, var, (score-mean)/(var+1.0e-10), ok, shuffle );
	      t=seq->s;
	      while( t < &seq->s[clip_seq] )
		putc(*(t++),stdout);
	      putc(' ',stdout);
	      printf("\n");
	      
	      free_seq(seq);
	    }
	}
      if ( sfp = argfile("-screen=%s","w",argc,argv,screenfile) ) /* write out a fasta file with the vector X'd out */
	{
	  printf("! writing fasta file %s\n");
	  clip_seq--;
	  fprintf(stderr, ">%s\n", screenfile );
	  if ( scf )
	    {
	      Header *h = &scf->header;
	      Bases *b = scf->bases;

	      for(i=0;i< h->bases;i++)
		{
		  if ( i <= clip_seq )
		    fputc('X',sfp);
		  else
		    fputc(b[i].base,sfp);
		  if ( i > 0 && ! (i%60) )
		    fputc('\n',sfp);
		}
	      fputc('\n',sfp);
	    }
	  else
	    {
	      for(i=0;i<seq->len;i++)
		{
		  if ( i <= clip_seq )
		    fputc('X',sfp);
		  else
		    fputc(seq->s[i],sfp);
		  if ( i > 0 && ! (i%60) )
		    fputc('\n',sfp);
		}
	      fputc('\n',sfp);
	    }
	  fclose(sfp);
	}
      return 0;
    }
  else
    return 1;
    }

SEQUENCE *get_5_vec( SEQUENCE *seq, int sc, int sp )
{
/* fixup a vector sequence to get the 5' sequence for matching to a reading 

sc is the cloning site in the sequencing vector
sp is the primer offset from the cloning site 

*/

  SEQUENCE *vec = (SEQUENCE*)calloc(1,sizeof(SEQUENCE));
  char *s, *t;
  int n, len;

  vec->name = strdup(seq->name);
  if ( sp > 0 ) /* forward primer */
    {
      vec->s = (char*)calloc(sp+2,sizeof(char));
      len = strlen(seq->s);
      for(n=0;n<=sp;n++)
	vec->s[n] = seq->s[(sc+n)%len];
      vec->len = strlen(vec->s);
      complement_seq(vec->s);
      printf("! vec %s\n", vec->s);
      return vec;
    }
  else /* reverse primer */
    {
      sp = -sp;
      vec->s = (char*)calloc(sp+2,sizeof(char));
      len = strlen(seq->s);
      for(n=0;n<=sp;n++)
	vec->s[n] = seq->s[(sc-sp+n+len)%len];
      vec->len = strlen(vec->s);
      printf("! vec %s\n", vec->s);
      return vec;
    }      
  return NULL;
}

SEQUENCE *get_3_vec( SEQUENCE *seq, int sc, int Len )
{
/* fixup a vector sequence to get the 3' sequence for matching to a reading 

sc is the cloning site in the sequencing vector
Len is the max length to align

*/

  SEQUENCE *vec = (SEQUENCE*)calloc(1,sizeof(SEQUENCE));
  char *s, *t;
  int n, len;

  vec->name = strdup(seq->name);
  len = strlen(seq->s);
  if ( Len > 0 ) /* forward primer */
    {
      if ( Len > len )
	Len = len;
      vec->s = (char*)calloc(Len+1,sizeof(char));
      for(n=0;n<Len;n++)
	vec->s[Len-n-1] = seq->s[(len+sc-n)%len];
      vec->len = strlen(vec->s);
/*      complement_seq(vec->s);    */
      printf("! 3' %d %d vec %s\n", Len, len, vec->s);
    }
  else /* reverse primer */
    {
      Len = -Len;
      if ( Len > len )
	Len = len;
      vec->s = (char*)calloc(Len+1,sizeof(char));
      len = strlen(seq->s);
      for(n=0;n<Len;n++)
	vec->s[n] = seq->s[(sc+n+len)%len];
      vec->len = strlen(vec->s);
      complement_seq(vec->s);   
/*      printf("! vec %d %s\n", Len, vec->s); */
    }      
  return vec;

}

void end_align( SEQUENCE *query, SEQUENCE *seq, 
	       int match, int mismatch, int gap, int max, int align_mode, 
	       int *q, int *s, int *score, int *matches )

/* 
   returns the Needleman Wunsch score for aligning the end of query to the begining of seq 
   matches score +match, mismatches +mismatch, gaps +gap (no gap start penalty)
*/
  
{
  int len;
  int *A, *B, *a, *b, *c;
  int *countA, *countB, *counta, *countb;
  int i1, i2;
  int max_score=-1000000;
  int c1=0, c2=0;
  int m;
  int d, u, l;
  int start1, stop1;
  int start2, stop2;
  int x2;
  char *lower;
  static int mat_init_ok=0;
  int match_count=0;
  char pad_char = '*';

  if ( ! mat_init_ok )
    {
      mat_init(match,mismatch,gap,0,pad_char);
      mat_init_ok = 1;
    }

  A = (int*)calloc(query->len+1,sizeof(int));
  B = (int*)calloc(query->len+1,sizeof(int));

  countA = (int*)calloc(query->len+1,sizeof(int));
  countB = (int*)calloc(query->len+1,sizeof(int));

  lower = (char*)downcase(strdup(query->s));
  
  a = A+1;
  b = B+1;

  counta = countA+1;
  countb = countB+1;

  start2 = 0;
  stop2 = seq->len-1;
  if ( max > 0 && stop2 > max )
    stop2 = max;

  for(i2=start2;i2<=stop2;i2++)
    {
      c = a;
      a = b;
      b = c;

      c = counta;
      counta = countb;
      countb = c;

      start1 = 0;
      stop1 = query->len-1;

      x2 = tolower(seq->s[i2]);

      for(i1=start1;i1<=stop1;i1++)
	{
	  d = a[i1-1] + (m=sim_mat[lower[i1]][x2]);
	  u = a[i1] + gap;
	  l = b[i1-1] + gap;
	  if ( d > u )
	    {
	      if ( d > l )
		{
		  b[i1] = d;
		  countb[i1] = (m>0) + counta[i1-1];
		}
	      else
		{
		  b[i1] = l;
		  countb[i1] = countb[i1-1];
		}
	    }
	  else if ( u > l )
	    {
	      b[i1] = u;
	      countb[i1] = counta[i1];
	    }
	  else
	    {
	      b[i1] = l;
	      countb[i1] = countb[i1-1];
	    }
	  if (  align_mode== SMITH_WATERMAN && b[i1] < 0 ) /* SW */
	    {
	      b[i1] = 0;     /* SW */
	      countb[i1] = 0;
	    }
	}
      i1--;
      if ( b[i1] > max_score )
	{
	  max_score = b[i1];
	  match_count = countb[i1];
	  c1 = i1;
	  c2 = i2;
	}
    }

  free(A);
  free(B);
  free(countA);
  free(countB);

  free(lower);

  *q = c1;
  *s = c2;
  *score = max_score;
  *matches = match_count;
}

void try_trace_alignment( Scf *scf, SEQUENCE *vec, int argc, char **argv, int *clip_point, int *score, float *mean, float *var, int *ok, int *shuffle )
{

  Header *h;
  int bases;
  int seq_start, seq_end, trace_start, trace_end;
  int *coord, *w_score, *w_max, *t_score, *t_max;
  int base_start, base_end, t_start, t_end;
  int s;
  float floor=.90;
  float t_weight=2;
  int align_mode=NEEDLEMAN;
  char mode[256];
  int max=70;
  int min_trace=7;
  int seed;
  int n, s_score, ss, s_matches;
  SEQUENCE *shuffled_vec;
  FILE *ps;
  char ps_file[256];
  int window_band=0;
  t_params params;
  int w, wm;
  int rescore=0;

  getint("-max=%d",&max,argc,argv); /* max clip in sequence */

  if ( *shuffle > 0 )
    {
      getseed(&seed,argc,argv);
      shuffled_vec = seqdup(vec);
    }

  getint("-min=%d",&min_trace,argc,argv);
  getfloat("-floor=%f",&floor,argc,argv);
  getfloat("t_weight=%f",&t_weight,argc,argv);

  strcpy(mode,"needleman");
  if ( getarg("-mode=%s", mode, argc,argv ) )
    legal_string( mode, align_modes, 2, &align_mode );

  params.minstate = min_trace;
  params.floor   = floor;
  params.deriv_weight = t_weight;
  params.mode = align_mode;
  params.samples_per_base = window_band;
  params.gap_penalty=0;

  h = &scf->header;
  bases = vec->len;
  trace_start = 0;
  if ( h->bases < max )
    trace_end = h->samples;
  else
    trace_end = scf->bases[max-1].peak_index;
  
  coord = (int*)calloc(bases+1,sizeof(int));
  w_score = (int*)calloc(bases+1,sizeof(int));
  w_max = (int*)calloc(bases+1,sizeof(int));
  t_score = (int*)calloc(bases+1,sizeof(int));
  t_max = (int*)calloc(bases+1,sizeof(int));
  
  trace_align( scf, trace_start, trace_end, vec->s, 0, vec->len-1, &params, rescore, coord, w_score, score, &base_start, &base_end, &t_start, &t_end );

  if ( ps = argfile("-ps=%s","w",argc,argv,ps_file) )
    {
      ps_print_alignment( ps, scf, trace_start, trace_end, base_end-base_start+1, vec->s, base_start, base_end, coord, ps_file );
      fclose(ps);
    }
  
  for(s=0;s<h->bases;s++)
    if ( scf->bases[s].peak_index > t_end )
      break;

  *clip_point = s;

  
  *ok = 0;
  *mean = *var = 0.0;

  wm = 0;
  w = 0;
  for(s=base_start;s<=base_end;s++)
    {
      wm += w_max[s];
      w += w_score[s];
    }
  printf( "! len: %d, w: %d wm %d  %.3f\n",  base_end-base_start+1, w, wm, 0.90*wm );
  if ( base_end-base_start+1 > 30 && w > 0.90*wm ) /* very strong similarity - no need to shuffle */
    *ok = *shuffle;
/* very weak - obviously no good */
/*  else if ( w < 0.5 *wm ) 
    *ok = 0; */
  else if ( *shuffle  ) /* do shuffles to test */
    {
      for(n=0;n<*shuffle;n++)
	{
	  shuffle_seq(shuffled_vec,1,&seed);
	  trace_align( scf, trace_start, trace_end, shuffled_vec->s, 0, shuffled_vec->len-1, &params, rescore, coord, w_score, &s_score, &base_start, &base_end, &t_start, &t_end );

	  *mean += s_score;
	  *var += s_score*s_score;
	  *ok += ( *score > s_score );
	}
      *mean /= *shuffle;
      *var = (*var-*shuffle*(*mean)*(*mean))/(*shuffle-1.0+1.0e-5);
      *var = sqrt(*var)+1.0e-5;
      free_seq(shuffled_vec);
    }
  

  free(coord);
  free(w_score);
  free(w_max);
  free(t_score);
  free(t_max);
}

void try_seq_alignment( SEQUENCE *seq, SEQUENCE *vec, 
		       int max_clip,
		       int argc, char **argv, int *clip_pos, 
		       int *clip_pos_vec, int *score, int *matches, 
		       float *mean, float *var, int *ok, int *shuffle )

/* try to match the vector to a sequence using a neddleman-type
alignment.  If a match of over 30 bases (or to whatever min_matches is
set) is found then this is treated as definitely correct.  Otherwise
the alignment is tested by repeated shuffling of the vector sequence.

The sequences should be input in such a way that we are always
matching the end of the vector to the start of the query.  This is
natural for 5' end, but you need to reverse-complement for the 3' end */

{
  SEQUENCE *shuffled_vec;
  int match=3, mismatch=-5, gap=-5;
  int n;
  int min_matches=30; /* min no of matches needed for a definite alignment */
  int align_mode=NEEDLEMAN;
  char mode[256];
  int seed;
  int q, s;
  int s_score, ss, s_matches;
  char *t;

  getint("-match=%d",&match,argc,argv);
  getint("-mismatch=%d",&mismatch,argc,argv);
  getint("-gap=%d",&gap,argc,argv);
  getint("-minmatch=%d",&min_matches,argc,argv); 
  
  if ( mismatch  > 0 )
    mismatch = -mismatch;
  if ( gap > 0 )
    gap = -gap;
  
  if ( *shuffle > 0 )
    {
      getseed(&seed,argc,argv);
      shuffled_vec = seqdup(vec);
    }

  strcpy(mode,"needleman");
  if ( getarg("-mode=%s", mode, argc,argv ) )
    legal_string( mode, align_modes, 2, &align_mode );

  *mean=0.0;
  *var = 0.0;
  *ok = 0;

  printf("!seq: %s\n", seq->s);
  printf("!vec: %s\n", vec->s);

  end_align( vec, seq, match, mismatch, gap, max_clip, align_mode, &q, &s, score, matches);
  printf("! score %d matches %d\n", *score, *matches ); 
  *clip_pos = s+1;
  *clip_pos_vec = q+1;

  if ( *matches > 30 ) /* an alignment with > 30 matches should be correct, so don't shuffle */
    *ok = ( *shuffle ? *shuffle : 1 );
  else if ( *shuffle )
    {
      printf("! shuffling %d\n", *shuffle);
      for(n=0;n<*shuffle;n++)
	{
	  shuffle_seq(shuffled_vec,1,&seed);
	  end_align( shuffled_vec, seq, match, mismatch, gap, max_clip, align_mode, &q, &ss, &s_score, &s_matches );
/*	  printf("! %5d score %d\n", n, s_score ); */
	  (*mean) += s_score;
	  (*var) += s_score*s_score;
	  *ok += ( (*score) > s_score );
	}
      (*mean) /= *shuffle;
      (*var) = ((*var)-*shuffle*(*mean)*(*mean))/(*shuffle-1.0+1.0e-5);
      (*var) = sqrt((*var))+1.0e-5;
    }
}

SEQUENCE *scf2seq( Scf *scf, char *name )/* extract the sequence from the scf */
{
  Header *h = &scf->header;
  SEQUENCE *seq = (SEQUENCE*)calloc(1,sizeof(SEQUENCE));
  int i;

  seq->len = h->bases;
  seq->s = (char*)calloc(seq->len+1,sizeof(char)); 
  for(i=0;i<seq->len;i++)
    seq->s[i] = scf->bases[i].base;
  seq->name = (char*)strdup(name);

  return seq;
}






